/*  -*- C++ -*- */

%skeleton "lalr1.cc"
%define "parser_class_name" "ValueExpressionParser"

%defines
%{
#include <volFields.H>
#include <fvcGrad.H>
#include <fvcDiv.H>

#include <functional>
#include <cmath>

#include "ValueExpressionDriver.H"
#include "ValueExpressionDriverLogicalTemplates.H"
%}

%name-prefix="ve"

%parse-param { ValueExpressionDriver& driver }
%lex-param { ValueExpressionDriver& driver }

%locations
%initial-action
{
	     // Initialize the initial location.
	     //     @$.begin.filename = @$.end.filename = &driver.file;
};

%debug
%error-verbose

%union
{
    Foam::scalar val;
    Foam::vector *vec;
    Foam::string *name;
    Foam::string *vname;
    Foam::volVectorField *vfield;
    Foam::volScalarField *sfield;
};

%token <name>   TOKEN_SID   "scalarID"
%token <vname>  TOKEN_VID   "vectorID"
%token <val>    TOKEN_NUM  "number"
%token <vec>    TOKEN_VEC  "vector"
%type  <sfield>    exp        "expression"
%type  <sfield>    lexp       "lexpression"
%type  <vfield>    vexp       "vexpression"
%type  <vfield>    vector     

%token TOKEN_VECTOR

%token TOKEN_TRUE
%token TOKEN_FALSE

%token TOKEN_pi
%token TOKEN_rand
%token TOKEN_randNormal
%token TOKEN_position
%token TOKEN_dist

%token TOKEN_div
%token TOKEN_grad

%token TOKEN_pow
%token TOKEN_log
%token TOKEN_exp
%token TOKEN_mag
%token TOKEN_sin
%token TOKEN_cos
%token TOKEN_tan
%token TOKEN_min
%token TOKEN_max
%token TOKEN_sqr
%token TOKEN_sqrt

%left '?' ':'
%left TOKEN_OR
%left TOKEN_AND
%left TOKEN_EQ TOKEN_NEQ
%left TOKEN_LEQ TOKEN_GEQ '<' '>'
%left '-' '+'
%left '*' '/' '&' '^'
%left TOKEN_NEG TOKEN_NOT
// %right '^'
%left '.'

%printer             { debug_stream () << *$$; } "scalarID" "vectorID"
%printer             { Foam::OStringStream buff; buff << *$$; debug_stream () << buff.str().c_str(); } "vector"
%destructor          { delete $$; } "scalarID" "vectorID" "vector" "expression" "vexpression" "lexpression"
%printer             { debug_stream () << $$; } "number" 
%printer             { debug_stream () << $$->name().c_str(); } "expression"  "vexpression" "lexpression"


%%
%start unit;

unit:   exp                     { driver.setScalarResult($1);  }
        | vexp                  { driver.setVectorResult($1);  }
        | lexp                  { driver.setLogicalResult($1); }
;

vexp:   vector                  { $$ = $1; }
        | vexp '+' vexp 		{ $$ = new Foam::volVectorField(*$1 + *$3); delete $1; delete $3; }
        | exp '*' vexp 		        { $$ = new Foam::volVectorField(*$1 * *$3); delete $1; delete $3; }
        | vexp '*' exp 		        { $$ = new Foam::volVectorField(*$1 * *$3); delete $1; delete $3; }
        | vexp '/' exp 		        { $$ = new Foam::volVectorField(*$1 / *$3); delete $1; delete $3; }
        | vexp '^' vexp 		{ $$ = new Foam::volVectorField(*$1 ^ *$3); delete $1; delete $3; }
        | vexp '-' vexp 		{ $$ = new Foam::volVectorField(*$1 - *$3); delete $1; delete $3;}
        | '-' vexp %prec TOKEN_NEG 	        { $$ = new Foam::volVectorField(-*$2); delete $2; }
        | '(' vexp ')'		        { $$ = $2; }  
        | lexp '?' vexp ':' vexp        { $$ = driver.doConditional($1,$3,$5,driver.makeVectorField(Foam::vector())); delete $1; delete $3; delete $5; }
        | TOKEN_position '(' ')'        { $$ = driver.makePositionField(); }
        | TOKEN_grad '(' exp ')'        { $$ = new Foam::volVectorField(Foam::fvc::grad(*$3)); delete $3; }
        | TOKEN_VID {
            $$=driver.getVectorField(*$1);
                    }
;

exp:	TOKEN_NUM		{ $$ = driver.makeScalarField($1); }
        | exp '+' exp 		{ $$ = new Foam::volScalarField(*$1 + *$3); delete $1; delete $3; }
        | exp '-' exp 		{ $$ = new Foam::volScalarField(*$1 - *$3); delete $1; delete $3; }
        | exp '*' exp 		{ $$ = new Foam::volScalarField(*$1 * *$3); delete $1; delete $3; }
        | exp '/' exp 		{ $$ = new Foam::volScalarField(*$1 / *$3); delete $1; delete $3; }
        | TOKEN_pow '(' exp ',' TOKEN_NUM ')'		{ $$ = new Foam::volScalarField(Foam::pow(*$3, $5)); delete $3; }
        | TOKEN_log '(' exp ')'       { $$ = new Foam::volScalarField(Foam::log(*$3)); delete $3; }
        | TOKEN_exp '(' exp ')'       { $$ = new Foam::volScalarField(Foam::exp(*$3)); delete $3; }
        | vexp '&' vexp 	{ $$ = new Foam::volScalarField(*$1 & *$3); delete $1; delete $3;}
        | '-' exp %prec TOKEN_NEG 	{ $$ = new Foam::volScalarField(-*$2); delete $2; }
	| '(' exp ')'		{ $$ = $2; }
        | TOKEN_sqr '(' exp ')'       { $$ = new Foam::volScalarField(Foam::sqr(*$3)); delete $3; }
        | TOKEN_sqrt '(' exp ')'       { $$ = new Foam::volScalarField(Foam::sqrt(*$3)); delete $3; }
        | TOKEN_sin '(' exp ')'       { $$ = new Foam::volScalarField(Foam::sin(*$3)); delete $3; }
        | TOKEN_cos '(' exp ')'       { $$ = new Foam::volScalarField(Foam::cos(*$3)); delete $3; }
        | TOKEN_tan '(' exp ')'       { $$ = new Foam::volScalarField(Foam::tan(*$3)); delete $3; }
        | TOKEN_min '(' exp ')'       { $$ = driver.makeScalarField(Foam::min(*$3).value()); delete $3; }
        | TOKEN_max '(' exp ')'       { $$ = driver.makeScalarField(Foam::max(*$3).value()); delete $3; }
        | TOKEN_mag '(' exp ')'       { $$ = new Foam::volScalarField(Foam::mag(*$3)); delete $3; }
        | TOKEN_mag '(' vexp ')'      { $$ = new Foam::volScalarField(Foam::mag(*$3)); delete $3; }
        | TOKEN_div '(' vexp ')'      { $$ = new Foam::volScalarField(Foam::fvc::div(*$3)); delete $3; }
        | vexp '.' 'x'            { $$ = new Foam::volScalarField($1->component(0)); delete $1; }
        | vexp '.' 'y'            { $$ = new Foam::volScalarField($1->component(1)); delete $1; }
        | vexp '.' 'z'            { $$ = new Foam::volScalarField($1->component(2)); delete $1; }
        | lexp '?' exp ':' exp        { $$ = driver.doConditional($1,$3,$5,driver.makeScalarField(0.)); delete $1; delete $3; delete $5; }
        | TOKEN_pi { $$ = driver.makeScalarField(M_PI); }
        | TOKEN_dist '(' ')'        { $$ = driver.makeDistanceField(); }
        | TOKEN_rand '(' ')'        { $$ = driver.makeRandomField(); }
        | TOKEN_randNormal '(' ')'        { $$ = driver.makeGaussRandomField(); }
	| TOKEN_SID		{ 
            $$=driver.getScalarField(*$1);
				}
;

lexp: TOKEN_TRUE   { $$ = driver.makeScalarField(1); }
    | TOKEN_FALSE  { $$ = driver.makeScalarField(0); }
    | exp '<' exp  { $$ = driver.doCompare($1,std::less<Foam::scalar>(),$3);  delete $1; delete $3; }
    | exp '>' exp  { $$ = driver.doCompare($1,std::greater<Foam::scalar>(),$3);  delete $1; delete $3; }
    | exp TOKEN_LEQ exp  { $$ = driver.doCompare($1,std::less_equal<Foam::scalar>(),$3);  delete $1; delete $3; }
    | exp TOKEN_GEQ exp  { $$ = driver.doCompare($1,std::greater_equal<Foam::scalar>(),$3);  delete $1; delete $3; }
    | exp TOKEN_EQ exp  { $$ = driver.doCompare($1,std::equal_to<Foam::scalar>(),$3);  delete $1; delete $3; }
    | exp TOKEN_NEQ exp  { $$ = driver.doCompare($1,std::not_equal_to<Foam::scalar>(),$3);  delete $1; delete $3; }
    | '(' lexp ')'		{ $$ = $2; }
    | lexp TOKEN_AND lexp  { $$ = driver.doLogicalOp($1,std::logical_and<Foam::scalar>(),$3);  delete $1; delete $3; }
    | lexp TOKEN_OR lexp   { $$ = driver.doLogicalOp($1,std::logical_or<Foam::scalar>(),$3);  delete $1; delete $3; }
    | '!' lexp %prec TOKEN_NOT { $$ = driver.doLogicalNot($2); delete $2; }
;

vector: TOKEN_VECTOR '(' exp ',' exp ',' exp ')' {     $$ = driver.makeVectorField($3,$5,$7);  delete $3; delete $5; delete $7}
;

%%

void yy::ValueExpressionParser::error (const yy::ValueExpressionParser::location_type& l,const std::string& m)
{
     driver.error (l, m);
}
